#ifndef AST_NODE_DATA_TYPE_HPP
#define AST_NODE_DATA_TYPE_HPP

#include <utils/PugsAssert.hpp>

#include <limits>
#include <memory>
#include <string>

class ASTNode;
class ASTNodeDataType;

ASTNodeDataType getVectorDataType(const ASTNode& type_node);

std::string dataTypeName(const ASTNodeDataType& data_type);

ASTNodeDataType dataTypePromotion(const ASTNodeDataType& data_type_1, const ASTNodeDataType& data_type_2);

bool isNaturalConversion(const ASTNodeDataType& data_type, const ASTNodeDataType& target_data_type);

class ASTNodeDataType
{
 public:
  enum DataType : int32_t
  {
    undefined_t        = -1,
    bool_t             = 0,
    int_t              = 1,
    unsigned_int_t     = 2,
    double_t           = 3,
    vector_t           = 4,
    tuple_t            = 5,
    list_t             = 6,
    string_t           = 7,
    typename_t         = 10,
    type_name_id_t     = 11,
    type_id_t          = 21,
    function_t         = 22,
    builtin_function_t = 23,
    void_t             = std::numeric_limits<int32_t>::max()
  };

 private:
  DataType m_data_type;
  std::shared_ptr<ASTNodeDataType> m_content_type;
  size_t m_dimension;
  std::string m_name_of_type_id;

 public:
  PUGS_INLINE
  size_t
  dimension() const
  {
    return m_dimension;
  }

  PUGS_INLINE
  const ASTNodeDataType&
  contentType() const
  {
    Assert(m_content_type);
    return *m_content_type;
  }

  PUGS_INLINE
  const std::string&
  nameOfTypeId() const
  {
    return m_name_of_type_id;
  }

  PUGS_INLINE
  operator const DataType&() const
  {
    return m_data_type;
  }

  ASTNodeDataType& operator=(const ASTNodeDataType&) = default;
  ASTNodeDataType& operator=(ASTNodeDataType&&) = default;

  template <DataType data_type>
  static ASTNodeDataType
  build()
  {
    static_assert(data_type != tuple_t, "tuple_t requires sub_type");
    static_assert(data_type != typename_t, "typename_t requires sub_type");
    static_assert(data_type != vector_t, "vector_t requires dimension");
    static_assert(data_type != type_id_t, "type_id_t requires name");

    return ASTNodeDataType{data_type};
  }

  template <DataType data_type>
  static ASTNodeDataType
  build(const ASTNodeDataType& content_type)
  {
    static_assert((data_type == tuple_t) or (data_type == typename_t),
                  "incorrect data_type construction: cannot have content");

    return ASTNodeDataType{data_type, content_type};
  }

  template <DataType data_type>
  static ASTNodeDataType
  build(size_t dimension)
  {
    static_assert((data_type == vector_t), "incorrect data_type construction: cannot have dimension");
    return ASTNodeDataType{data_type, dimension};
  }

  template <DataType data_type>
  static ASTNodeDataType
  build(const std::string& type_name)
  {
    static_assert((data_type == type_id_t), "incorrect data_type construction: cannot provide name of type");
    Assert(type_name != "__undefined_name_type");

    return ASTNodeDataType{data_type, type_name};
  }

  ASTNodeDataType()
    : m_data_type{undefined_t}, m_content_type{nullptr}, m_dimension{1}, m_name_of_type_id{"__undefined_name_type"}
  {}

  ASTNodeDataType(const ASTNodeDataType&) = default;

  ASTNodeDataType(ASTNodeDataType&&) = default;

  ~ASTNodeDataType() = default;

 private:
  explicit ASTNodeDataType(DataType data_type)
    : m_data_type{data_type}, m_content_type{nullptr}, m_dimension{1}, m_name_of_type_id{"__undefined_name_type"}
  {}

  explicit ASTNodeDataType(DataType data_type, const ASTNodeDataType& content_type)
    : m_data_type{data_type},
      m_content_type{std::make_shared<ASTNodeDataType>(content_type)},
      m_dimension{1},
      m_name_of_type_id{"__undefined_name_type"}
  {}

  explicit ASTNodeDataType(DataType data_type, size_t dimension)
    : m_data_type{data_type},
      m_content_type{nullptr},
      m_dimension{dimension},
      m_name_of_type_id{"__undefined_name_type"}
  {}

  explicit ASTNodeDataType(DataType data_type, const std::string& type_name)
    : m_data_type{data_type}, m_content_type{nullptr}, m_dimension{1}, m_name_of_type_id{type_name}
  {}
};

#endif   // AST_NODE_DATA_TYPE_HPP
